/******************************************************************************
 *  CVS version:
 *     $Id: $
 ******************************************************************************
 *
 *  C code file : general.c
 *  Project     : Alan Compiler
 *  Version     : 1.0 alpha
 *  Description : Generic definitions
 *
 *  Comments: (in Greek iso-8859-7)
 *  ---------
 *    .
 *       .
 *      .
 *    
 */


/* ---------------------------------------------------------------------
   ---------------------------- Header files ---------------------------
   --------------------------------------------------------------------- */

#include <stdlib.h>
#include <string.h>

#if USEGC
#include "gc.h"
#endif

#include "general.h"
#include "error.h"


/* ---------------------------------------------------------------------
   --------------------------- Generic flags ---------------------------
   --------------------------------------------------------------------- */

bool flagVerbose = true;


/* ---------------------------------------------------------------------
   -------------     -------------
   --------------------------------------------------------------------- */

void terminate (int code)
{
   if (flagVerbose)
      message("\nErrors:   %d\nWarnings: %d", numErrors, numWarnings);
   exit(code);
}


/* ---------------------------------------------------------------------
   -----------      ------------
   --------------------------------------------------------------------- */

void * new (size_t size)
{
   void * result =
#if USEGC
      GC_MALLOC(size);
#else
   malloc(size);
#endif

   if (result == NULL)
      fatal("\rOut of memory");
   return result;
}

void delete (void * p)
{
   if (p != NULL)
#if USEGC
      GC_FREE(p);
#else
   free(p);
#endif
}


/* ---------------------------------------------------------------------
   -------        --------
   --------------------------------------------------------------------- */

char * filename;
int linecount;


/* ---------------------------------------------------------------------
   ------------     ------------
   --------------------------------------------------------------------- */

typedef unsigned long int HashType;

static HashType PJW_hash (const char * key)
{
   /*
    *  P.J. Weinberger's hashing function. See also:
    *  Aho A.V., Sethi R. & Ullman J.D, "Compilers: Principles,
    *  Techniques and Tools", Addison Wesley, 1986, pp. 433-437.
    */

   const HashType PJW_OVERFLOW =
      (((HashType) 0xf) << (8 * sizeof(HashType) - 4));
   const int PJW_SHIFT = (8 * (sizeof(HashType) - 1));

   HashType h, g;

   for (h = 0; *key != '\0'; key++) {
      h = (h << 4) + (*key);
      if ((g = h & PJW_OVERFLOW) != 0) {
         h ^= g >> PJW_SHIFT;
         h ^= g;
      }
   }
   return h;
}


/* ---------------------------------------------------------------------
   ---------------------------  ---------------------------
   --------------------------------------------------------------------- */

struct Identifier_tag {
      char * name;
      Identifier next;
};

static Identifier id_make_new (const char * name, Identifier next)
{
   Identifier result = new (sizeof (struct Identifier_tag));

   result->name = new(strlen(name) + 1);
   strcpy(result->name, name);
   result->next = next;
   return result;
}

/*           */

#define SIZE 193

/*      (static) global 
   :    hashtable[i] == NULL   i.
           id_reset. */

static Identifier hashtable [SIZE];

Identifier id_make (const char * name)
{
   HashType index = PJW_hash(name) % SIZE;

   Identifier ids = hashtable[index];
   Identifier id;

   for (id=ids; id != NULL; id=id->next)
      if (strcmp(id->name, name) == 0)
         return id;
   id = id_make_new(name, ids);
   hashtable[index] = id;
   return id;
}

const char * id_name (Identifier id)
{
   return id->name;
}

/* :         
       . */

void id_reset ()
{
   int i;

   for (i=0; i<SIZE; i++)
      hashtable[i] = NULL;
}
