/******************************************************************************
 *  CVS version:
 *     $Id: $
 ******************************************************************************
 *
 *  C header file : symbol.h
 *  Project       : Alan Compiler
 *  Version       : 1.0 alpha
 *  Description   : Generic symbol table in C
 *
 *  Comments: (in Greek iso-8859-7)
 *  ---------
 *    .
 *       .
 *      .
 *    
 */


#ifndef __SYMBOL_H__
#define __SYMBOL_H__


/* ---------------------------------------------------------------------
   ---------------------------- Header files ---------------------------
   --------------------------------------------------------------------- */

#include <stdbool.h>

#include "general.h"
#include "types.h"


/* ---------------------------------------------------------------------
   ------------      -------------------
   --------------------------------------------------------------------- */

#define START_POSITIVE_OFFSET 8     /*   offset  ..   */
#define START_NEGATIVE_OFFSET 0     /*   offset  .. */


/* ---------------------------------------------------------------------
   ---------------      -------------------
   --------------------------------------------------------------------- */

/*      */

typedef enum {            
   ENTRY_VARIABLE,                       /*                  */
   ENTRY_FUNCTION,                       /*                 */
   ENTRY_PARAMETER,                      /*       */
   ENTRY_TEMPORARY                       /*        */
} EntryType;


/*    */

typedef enum {            
   PASS_BY_VALUE,                        /* '                   */
   PASS_BY_REFERENCE                     /* '                */
} PassMode;


/*        */

typedef struct SymbolEntry_tag SymbolEntry;
typedef struct Scope_tag Scope;


/*      */

struct SymbolEntry_tag {
   Identifier     id;                 /*            */
   EntryType      entryType;          /*               */
   Scope        * scope;              /*          */
   SymbolEntry  * nextInHash;         /*    ..     */
   SymbolEntry  * nextInScope;        /*     */

   union {                            /*     : */

      struct {                                /*******  *******/
         Type          type;                  /*                  */
         int           offset;                /* Offset  ..       */
      } eVariable;

      struct {                                /*******  *******/
         bool          isForward;             /*  forward        */
         SymbolEntry * firstArgument;         /*        */
         SymbolEntry * lastArgument;          /*    */
         Type          resultType;            /*     */
         enum {                               /*    */
             PARDEF_COMPLETE,                    /*       */
             PARDEF_DEFINE,                      /*       */
             PARDEF_CHECK                        /*       */
         } pardef;
         int           firstQuad;             /*          */
      } eFunction;

      struct {                                /******  *******/
         Type          type;                  /*                  */
         int           offset;                /* Offset  ..       */
         PassMode      mode;                  /*       */
         SymbolEntry * next;                  /*      */
      } eParameter;

      struct {                                /**   **/
         Type          type;                  /*                  */
         int           offset;                /* Offset  ..       */
         int           number;
      } eTemporary;

   } u;                               /*   union               */
};


/*         */

struct Scope_tag {
   unsigned int   nesting;                   /*        */
   unsigned int   negOffset;                 /*   offset */
   Scope        * parent;                    /*    */
   SymbolEntry  * entries;                   /*     */
};


/*      */

typedef enum {
   LOOKUP_CURRENT_SCOPE,
   LOOKUP_ALL_SCOPES
} LookupType;


/* ---------------------------------------------------------------------
   -------------      --------------
   --------------------------------------------------------------------- */

extern Scope        * currentScope;       /*           */
extern unsigned int   quadNext;           /*    */
extern unsigned int   tempNumber;         /*   temporaries  */


/* ---------------------------------------------------------------------
   ------        ------
   --------------------------------------------------------------------- */

void          initSymbolTable    (unsigned int size);
void          destroySymbolTable (void);

void          openScope          (void);
void          closeScope         (void);

SymbolEntry * newVariable        (Identifier id, Type type, bool err);
SymbolEntry * newFunction        (Identifier id, bool err);
SymbolEntry * newParameter       (Identifier id, Type type, PassMode mode,
                                  SymbolEntry * f, bool err);
SymbolEntry * newTemporary       (Type type);

void          forwardFunction    (SymbolEntry * f);
void          endFunctionHeader  (SymbolEntry * f, Type type);
void          destroyEntry       (SymbolEntry * e);
SymbolEntry * lookupEntry        (Identifier id, LookupType type, bool err);


#endif
