/******************************************************************************
 *  CVS version:
 *     $Id: $
 ******************************************************************************
 *
 *  C code file : symbtest.c
 *  Project     : Alan Compiler
 *  Version     : 1.0 alpha
 *  Description : Generic symbol table in C, demo program
 *
 *  Comments: (in Greek iso-8859-7)
 *  ---------
 *    .
 *       .
 *      .
 *    
 */


/* Header files */

#include <stdlib.h>
#include <stdio.h>

#include "symbol.h"


/*      */

#define SHOW_OFFSETS 0


void printType (Type type)
{
   if (type == NULL) {
      printf("<undefined>");
      return;
   }

   switch (type->kind) {
      case TYPE_int:
         printf("int");
         break;
      case TYPE_byte:
         printf("byte");
         break;
      case TYPE_array:
         printType(type->elType);
         if (type->size > 0)
            printf(" [%d]", type->size);
         else
            printf(" []");
         break;
      case TYPE_proc:
         printf("proc");
         break;
      default:
         printf("<illegal>");
   }
}

void printMode (PassMode mode)
{
   if (mode == PASS_BY_REFERENCE)
      printf("reference ");
}

void printSymbolTable ()
{
   Scope       * scp;
   SymbolEntry * e;
   SymbolEntry * args;

   scp = currentScope;
   if (scp == NULL)
      printf("no scope\n");
   else
      while (scp != NULL) {
         printf("scope: ");
         e = scp->entries;
         while (e != NULL) {
            if (e->entryType == ENTRY_TEMPORARY)
               printf("$%d", e->u.eTemporary.number);
            else
               printf("%s", id_name(e->id));
            switch (e->entryType) {
               case ENTRY_FUNCTION:
                  printf("(");
                  args = e->u.eFunction.firstArgument;
                  while (args != NULL) {
                     printMode(args->u.eParameter.mode);
                     printf("%s : ", id_name(args->id));
                     printType(args->u.eParameter.type);
                     args = args->u.eParameter.next;
                     if (args != NULL)
                        printf("; ");
                  }
                  printf(") : ");
                  printType(e->u.eFunction.resultType);
                  break;
#if SHOW_OFFSETS
               case ENTRY_VARIABLE:
                  printf("[%d]", e->u.eVariable.offset);
                  break;
               case ENTRY_PARAMETER:
                  printf("[%d]", e->u.eParameter.offset);
                  break;
               case ENTRY_TEMPORARY:
                  printf("[%d]", e->u.eTemporary.offset);
                  break;
#endif
               default:
                  break;
            }
            e = e->nextInScope;
            if (e != NULL)
               printf(", ");
         }
         scp = scp->parent;
         printf("\n");
      }
   printf("----------------------------------------\n");
}


/*       */

/*      Alan  
        .

   p () : proc -- this is the program header

   s1 : byte; s2 : byte; s3 : byte;
   i1 : int; i2 : int;

   pr (p1 : int, p2 : int, p3 : reference byte) : proc
   b1 : byte;
   i1 : int;
   { -- of procedure pr
   -- access s1, i2, i1
   -- make 2 new temporaries
   } -- of procedure pr

   f (x : int; reference y : byte) : int
   { -- of function f
   -- ... whatever ...
   } -- of function f

   { -- of program p
   (* make 2 new temporaries *)
   } -- of program p
*/

int main ()
{
   /*   */

   SymbolEntry * p;

   /*  ! */

   initSymbolTable(256);

   printSymbolTable();

   /* SOURCE : p () : proc -- this is the program header */

   openScope();

   printSymbolTable();

   /* SOURCE : s1 : byte; s2 : byte; s3 : byte; */

   newVariable(id_make("s1"), type_byte(), true);
   newVariable(id_make("s2"), type_byte(), true);
   newVariable(id_make("s3"), type_byte(), true);

   printSymbolTable();

   /* SOURCE : i1 : int; i2 : int; */

   newVariable(id_make("i1"), type_int(), true);
   newVariable(id_make("i2"), type_int(), true);

   printSymbolTable();

   /* SOURCE : pr (... */

   p = newFunction(id_make("pr"), true);
   openScope();

   printSymbolTable();

   /* SOURCE : p1 : int, p2 : int, p3 : reference byte) : proc */

   newParameter(id_make("p1"), type_int(), PASS_BY_VALUE, p, true);
   newParameter(id_make("p2"), type_int(), PASS_BY_VALUE, p, true);
   newParameter(id_make("p3"), type_byte(), PASS_BY_REFERENCE, p, true);
   endFunctionHeader(p, type_proc());

   printSymbolTable();

   /* SOURCE : b1 : byte; */

   newVariable(id_make("b1"), type_byte(), true);

   /* SOURCE : i1 : int; */

   newVariable(id_make("i1"), type_int(), true);

   printSymbolTable();

   /* SOURCE : (* access s1, i2, i1 *) */

   lookupEntry(id_make("s1"), LOOKUP_ALL_SCOPES, true);
   lookupEntry(id_make("i2"), LOOKUP_ALL_SCOPES, true);
   lookupEntry(id_make("i1"), LOOKUP_ALL_SCOPES, true);

   /* SOURCE : (* make 2 new temporaries *) */

   newTemporary(type_int());
   newTemporary(type_byte());

   printSymbolTable();

   /* SOURCE : } -- of procedure pr */

   closeScope();

   printSymbolTable();

   /* SOURCE : f (... */

   p = newFunction(id_make("f"), true);
   openScope();

   printSymbolTable();

   /* SOURCE : x : int; reference y : byte) : int */

   newParameter(id_make("x"), type_int(), PASS_BY_VALUE, p, true);
   newParameter(id_make("y"), type_byte(), PASS_BY_REFERENCE, p, true);
   endFunctionHeader(p, type_int());

   printSymbolTable();

   /* SOURCE : } -- of procedure f */

   closeScope();

   printSymbolTable();

   /* SOURCE : (* make 2 new temporaries *) */

   newTemporary(type_int());
   newTemporary(type_int());

   printSymbolTable();

   /* SOURCE : } -- of program p */

   closeScope();

   printSymbolTable();

   return 0;
}
